// Julia version based on the simpler "imagehack". Allows modification of scale and number of iterations. Sets the offset by mouse movements.

// Important note: This implementation uses iterations, a for-loop, rather than recursion, since compute shaders can't do recursion, just like other shaders. The same thing holds for ray tracing, you need to reformulate the problem.

// As delivered, this works:
// g++ -o julia julia.cpp CSutilities.c ../common/Linux/MicroGlut.c ../common/*.c -I ../common/ -I ../common/Linux/ -DGL_GLEXT_PROTOTYPES -lm -lGL -lX11

#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include <GL/gl.h>
#define MAIN
#include "GL_utilities.h"
#include "LittleOBJLoader.h"
#include "MicroGlut.h"
#include "CSutilities.h"

GLfloat vertices[] = {	-1.0f,-1.0f,0.0f,
			-1.0f,1.0f,0.0f,
			1.0f,-1.0f,0.0f,
			1.0f,1.0f,0.0f };

GLfloat texcoord[] = {	0.0f, 0.0f,
			0.0f, 1.0f,
			1.0f, 0.0f,
			1.0f, 1.0f};
GLuint indicies[] = {0, 1, 2, 1, 2, 3};

Model *quad;
GLuint quadprogram;
GLuint myComputeShader;

GLfloat zoom = 2.0;
GLfloat time = 0.0;
GLuint maxIteration = 16;
GLfloat offx = 0.5, offy = 0.5;
char animation = 0;

void display(void)
{
	printError("pre display");

	if (animation)
		time += 0.1; // glutGet(GLUT_TIME);
	glUseProgram(myComputeShader);
	glUniform1f(glGetUniformLocation(myComputeShader, "time"), time);

	glDispatchCompute(32, 32, 1 );
	glMemoryBarrier( GL_ALL_BARRIER_BITS );

	glUseProgram(quadprogram);

	// clear the screen
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH);
	glClear(GL_COLOR_BUFFER_BIT);

	DrawModel(quad, quadprogram, "inPosition", NULL, "inTexCoord");
	
	printError("display");
	
//	glutSwapBuffers();
	glFlush();
}

void mouseMove(int x, int y)
{
	offx = (x-256) * 0.01;
	offy = (y-256) * 0.01;
	glUseProgram(myComputeShader);
	glUniform1f(glGetUniformLocation(myComputeShader, "offx"), offx);
	glUniform1f(glGetUniformLocation(myComputeShader, "offy"), offy);
	glUseProgram(quadprogram);
	glutPostRedisplay();
}

void keyDown(unsigned char key, int x, int y)
{
	if (key == '+')
		maxIteration *= 2;
	if (key == '-')
		if (maxIteration > 1)
			maxIteration /= 2;
	if (key == '.')
		zoom *= 1.5;
	if (key == ',')
		zoom /= 1.5;

	if (key == 'w')
		offy += 0.1;
	if (key == 'a')
		offx -= 0.1;
	if (key == 's')
		offy -= 0.1;
	if (key == 'd')
		offx += 0.1;
	if (key == ' ')
		animation = !animation;
		
	printf("zoom is %f\n", zoom);
	printf("maxIteration is %d\n", maxIteration);
	glUseProgram(myComputeShader);
	glUniform1f(glGetUniformLocation(myComputeShader, "zoom"), zoom);
	glUniform1f(glGetUniformLocation(myComputeShader, "offx"), offx);
	glUniform1f(glGetUniformLocation(myComputeShader, "offy"), offy);
	glUseProgram(quadprogram);
	glutPostRedisplay();
}

int main(int argc, char **argv)
{
// Let MicroGlut create a GL context
	glutInit (&argc, argv);
	glutInitContextVersion(4, 5);
	glutInitWindowSize(512, 512);
	glutCreateWindow("Julia CS");
	glutDisplayFunc(display);
	glutKeyboardFunc(keyDown);
	glutPassiveMotionFunc(mouseMove);
	glutRepeatingTimer(100);

// Let us know what we are running at.
	printf ("Vendor: %s\n", glGetString (GL_VENDOR));
	printf ("Renderer: %s\n", glGetString (GL_RENDERER));
	printf ("Version: %s\n", glGetString (GL_VERSION));
	printf ("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));

// Load and compile the compute shader    
	myComputeShader =loadComputeShader("julia.cs");

	GLuint tex;
	
	// Create texture
	
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, 512, 512, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glGenerateMipmap(GL_TEXTURE_2D);

	glBindImageTexture(0, tex, 0, GL_FALSE, 0, GL_READ_WRITE, GL_RGBA32F);
	printError("bind");

	// Send some default values to the shader	

	glUseProgram(myComputeShader);
	
	glUniform1i(glGetUniformLocation(myComputeShader, "tex"), 0);
	glUniform1f(glGetUniformLocation(myComputeShader, "zoom"), zoom);
	glUniform1f(glGetUniformLocation(myComputeShader, "offx"), offx);
	glUniform1f(glGetUniformLocation(myComputeShader, "offy"), offy);
	glUniform1i(glGetUniformLocation(myComputeShader, "maxIteration"), maxIteration);

	// Load the graphics shader and upload the geometry - a quad
	// I use LittleOBJLoader for convenience but it is really overkill.
	
	quadprogram = loadShaders("textured.vert", "textured.frag");
	glUseProgram(quadprogram);
	glUniform1i(glGetUniformLocation(quadprogram, "tex"), 0);
	
	quad = LoadDataToModel((vec3 *)vertices, NULL, (vec2 *)texcoord, NULL,  indicies,  4,  6);
	
	glutMainLoop();
}

